@@ -4,6 +4,8 @@
  * The redistribution terms are provided in the COPYRIGHT file that must
  * be distributed with this source code.
  */
+#include "config.h"
+
 #include <stdio.h>
 #include <string.h>
 #include <stdarg.h>
@@ -11,47 +13,17 @@
 #include <time.h>
 #include <ctype.h>
 
-#include "config.h"
-
 #include "cstring.h"
 #include "markdown.h"
 #include "amalloc.h"
-
-/* block-level tags for passing html blocks through the blender
- */
-struct kw {
-    char *id;
-    int  siz;
-} ;
-
-#define KW(x)	{ x, sizeof(x)-1 }
-
-static struct kw blocktags[] = { KW("!--"), KW("STYLE"), KW("SCRIPT"),
-				 KW("ADDRESS"), KW("BDO"), KW("BLOCKQUOTE"),
-				 KW("CENTER"), KW("DFN"), KW("DIV"), KW("H1"),
-				 KW("H2"), KW("H3"), KW("H4"), KW("H5"),
-				 KW("H6"), KW("LISTING"), KW("NOBR"),
-				 KW("UL"), KW("P"), KW("OL"), KW("DL"),
-				 KW("PLAINTEXT"), KW("PRE"), KW("TABLE"),
-				 KW("WBR"), KW("XMP"), KW("HR"), KW("BR") };
-#define SZTAGS	(sizeof blocktags / sizeof blocktags[0])
-#define MAXTAG	11 /* sizeof "BLOCKQUOTE" */
+#include "tags.h"
 
 typedef int (*stfu)(const void*,const void*);
 
 typedef ANCHOR(Paragraph) ParagraphRoot;
 
-
-/* case insensitive string sort (for qsort() and bsearch() of block tags)
- */
-static int
-casort(struct kw *a, struct kw *b)
-{
-    if ( a->siz != b->siz )
-	return a->siz - b->siz;
-    return strncasecmp(a->id, b->id, b->siz);
-}
-
+static Paragraph *Pp(ParagraphRoot *, Line *, int);
+static Paragraph *compile(Line *, int, MMIOT *);
 
 /* case insensitive string sort for Footnote tags.
  */
@@ -108,7 +80,7 @@ mkd_firstnonblank(Line *p)
 }
 
 
-static int
+static inline int
 blankline(Line *p)
 {
     return ! (p && (S(p->text) > p->dle) );
@@ -125,26 +97,35 @@ skipempty(Line *p)
 
 
 void
-___mkd_tidy(Line *t)
+___mkd_tidy(Cstring *t)
 {
-    while ( S(t->text) && isspace(T(t->text)[S(t->text)-1]) )
-	--S(t->text);
+    while ( S(*t) && isspace(T(*t)[S(*t)-1]) )
+	--S(*t);
 }
 
 
-static char *
+static struct kw comment = { "!--", 3, 0 };
+
+static struct kw *
 isopentag(Line *p)
 {
     int i=0, len;
-    struct kw key, *ret;
+    char *line;
 
     if ( !p ) return 0;
 
+    line = T(p->text);
     len = S(p->text);
 
-    if ( len < 3 || T(p->text)[0] != '<' )
+    if ( len < 3 || line[0] != '<' )
 	return 0;
 
+    if ( line[1] == '!' && line[2] == '-' && line[3] == '-' )
+	/* comments need special case handling, because
+	 * the !-- doesn't need to end in a whitespace
+	 */
+	return &comment;
+    
     /* find how long the tag is so we can check to see if
      * it's a block-level tag
      */
@@ -153,81 +134,191 @@ isopentag(Line *p)
 		       && !isspace(T(p->text)[i]); ++i )
 	;
 
-    key.id = T(p->text)+1;
-    key.siz = i-1;
-    
-    if ( ret = bsearch(&key,blocktags,SZTAGS,sizeof key, (stfu)casort))
-	return ret->id;
 
-    return 0;
+    return mkd_search_tags(T(p->text)+1, i-1);
 }
 
 
+typedef struct _flo {
+    Line *t;
+    int i;
+} FLO;
+
+#define floindex(x) (x.i)
+
+
 static int
-selfclose(Line *t, char *tag)
+flogetc(FLO *f)
 {
-    char *q = T(t->text);
-    int siz = strlen(tag);
-    int i;
+    if ( f && f->t ) {
+	if ( f->i < S(f->t->text) )
+	    return T(f->t->text)[f->i++];
+	f->t = f->t->next;
+	f->i = 0;
+	return flogetc(f);
+    }
+    return EOF;
+}
 
-    if ( strcasecmp(tag, "HR") == 0 || strcasecmp(tag, "BR") == 0 )
-	/* <HR> and <BR> are self-closing block-level tags,
-	 */
-	return 1;
 
-    i = S(t->text) - (siz + 3);
+static void
+splitline(Line *t, int cutpoint)
+{
+    if ( t && (cutpoint < S(t->text)) ) {
+	Line *tmp = calloc(1, sizeof *tmp);
+
+	tmp->next = t->next;
+	t->next = tmp;
+
+	tmp->dle = t->dle;
+	SUFFIX(tmp->text, T(t->text)+cutpoint, S(t->text)-cutpoint);
+	S(t->text) = cutpoint;
+    }
+}
+
+#define UNCHECK(l) ((l)->flags &= ~CHECKED)
+
+/*
+ * walk a line, seeing if it's any of half a dozen interesting regular
+ * types.
+ */
+static void
+checkline(Line *l)
+{
+    int eol, i;
+    int dashes = 0, spaces = 0,
+	equals = 0, underscores = 0,
+	stars = 0, tildes = 0;
+
+    l->flags |= CHECKED;
+    l->kind = chk_text;
+    l->count = 0;
+    
+    if (l->dle >= 4) { l->kind=chk_code; return; }
+
+    for ( eol = S(l->text); eol > l->dle && isspace(T(l->text)[eol-1]); --eol )
+	;
+
+    for (i=l->dle; i<eol; i++) {
+	register int c = T(l->text)[i];
+
+	if ( c != ' ' ) l->count++;
+
+	switch (c) {
+	case '-':  dashes = 1; break;
+	case ' ':  spaces = 1; break;
+	case '=':  equals = 1; break;
+	case '_':  underscores = 1; break;
+	case '*':  stars = 1; break;
+	case '~':  tildes = 1; break;
+	default:   return;
+	}
+    }
+
+    if ( dashes + equals + underscores + stars + tildes > 1 )
+	return;
+
+    if ( spaces ) {
+	if ( (underscores || stars || dashes) )
+	    l->kind = chk_hr;
+	return;
+    }
+
+    if ( stars || underscores ) { l->kind = chk_hr; }
+    else if ( dashes ) { l->kind = chk_dash; }
+    else if ( tildes ) { l->kind = chk_tilde; }
+    else if ( equals ) { l->kind = chk_equal; }
+}
+
+
+
+static Line *
+commentblock(Paragraph *p, int *unclosed)
+{
+    Line *t, *ret;
+    char *end;
+
+    for ( t = p->text; t ; t = t->next) {
+	if ( end = strstr(T(t->text), "-->") ) {
+	    splitline(t, 3 + (end - T(t->text)) );
+	    ret = t->next;
+	    t->next = 0;
+	    return ret;
+	}
+    }
+    *unclosed = 1;
+    return t;
 
-    /* we specialcase start and end tags on the same line.
-     */
-    return ( i > 0 ) && (q[i] == '<') && (q[i+1] == '/')
-		     && (q[i+2+siz] == '>')
-		     && (strncasecmp(&q[i+2], tag, siz) == 0);
 }
 
 
 static Line *
-htmlblock(Paragraph *p, char *tag)
+htmlblock(Paragraph *p, struct kw *tag, int *unclosed)
 {
-    Line *t = p->text, *ret;
-    int closesize;
-    char close[MAXTAG+4];
+    Line *ret;
+    FLO f = { p->text, 0 };
+    int c;
+    int i, closing, depth=0;
 
-    if ( selfclose(t, tag) || (strlen(tag) >= MAXTAG) ) {
-	ret = t->next;
-	t->next = 0;
+    *unclosed = 0;
+    
+    if ( tag == &comment )
+	return commentblock(p, unclosed);
+    
+    if ( tag->selfclose ) {
+	ret = f.t->next;
+	f.t->next = 0;
 	return ret;
     }
 
-    closesize = sprintf(close, "</%s>", tag);
+    while ( (c = flogetc(&f)) != EOF ) {
+	if ( c == '<' ) {
+	    /* tag? */
+	    c = flogetc(&f);
+	    if ( c == '!' ) { /* comment? */
+		if ( flogetc(&f) == '-' && flogetc(&f) == '-' ) {
+		    /* yes */
+		    while ( (c = flogetc(&f)) != EOF ) {
+			if ( c == '-' && flogetc(&f) == '-'
+				      && flogetc(&f) == '>')
+			      /* consumed whole comment */
+			      break;
+		    }
+		}
+	    }
+	    else { 
+		if ( closing = (c == '/') ) c = flogetc(&f);
 
-    for ( ; t ; t = t->next) {
-	if ( strncasecmp(T(t->text), close, closesize) == 0 ) {
-	    ret = t->next;
-	    t->next = 0;
-	    return ret;
+		for ( i=0; i < tag->size; c=flogetc(&f) ) {
+		    if ( tag->id[i++] != toupper(c) )
+			break;
+		}
+
+		if ( (i == tag->size) && !isalnum(c) ) {
+		    depth = depth + (closing ? -1 : 1);
+		    if ( depth == 0 ) {
+			while ( c != EOF && c != '>' ) {
+			    /* consume trailing gunk in close tag */
+			    c = flogetc(&f);
+			}
+			if ( c == EOF )
+			    break;
+			if ( !f.t )
+			    return 0;
+			splitline(f.t, floindex(f));
+			ret = f.t->next;
+			f.t->next = 0;
+			return ret;
+		    }
+		}
+	    }
 	}
     }
+    *unclosed = 1;
     return 0;
 }
 
 
-static Line *
-comment(Paragraph *p, char *key)
-{
-    Line *t, *ret;
-
-    for ( t = p->text; t ; t = t->next) {
-	if ( strstr(T(t->text), "-->") ) {
-	    ret = t->next;
-	    t->next = 0;
-	    return ret;
-	}
-    }
-    return t;
-
-}
-
-
 /* footnotes look like ^<whitespace>{0,3}[stuff]: <content>$
  */
 static int
@@ -241,88 +332,53 @@ isfootnote(Line *t)
     for ( ++i; i < S(t->text) ; ++i ) {
 	if ( T(t->text)[i] == '[' )
 	    return 0;
-	else if ( T(t->text)[i] == ']' && T(t->text)[i+1] == ':' )
-	    return 1;
+	else if ( T(t->text)[i] == ']' )
+	    return ( T(t->text)[i+1] == ':' ) ;
     }
     return 0;
 }
 
 
-static int
+static inline int
 isquote(Line *t)
 {
-    return ( T(t->text)[0] == '>' );
+    return (t->dle < 4 && T(t->text)[t->dle] == '>');
 }
 
 
-static int
-dashchar(char c)
-{
-    return (c == '*') || (c == '-') || (c == '_');
-}
-
-
-static int
+static inline int
 iscode(Line *t)
 {
     return (t->dle >= 4);
 }
 
 
-static int
+static inline int
 ishr(Line *t)
 {
-    int i, count=0;
-    char dash = 0;
-    char c;
+    if ( ! (t->flags & CHECKED) )
+	checkline(t);
 
-    if ( iscode(t) ) return 0;
-
-    for ( i = 0; i < S(t->text); i++) {
-	c = T(t->text)[i];
-	if ( (dash == 0) && dashchar(c) )
-	    dash = c;
-
-	if ( c == dash ) ++count;
-	else if ( !isspace(c) )
-	    return 0;
-    }
-    return (count >= 3);
+    if ( t->count > 2 )
+	return t->kind == chk_hr || t->kind == chk_dash || t->kind == chk_equal;
+    return 0;
 }
 
 
 static int
-ishdr(Line *t, int *htyp)
+issetext(Line *t, int *htyp)
 {
-    int i;
-
-
-    /* first check for etx-style ###HEADER###
+    Line *n;
+    
+    /* check for setext-style HEADER
+     *                        ======
      */
 
-    /* leading run of `#`'s ?
-     */
-    for ( i=0; T(t->text)[i] == '#'; ++i)
-	;
+    if ( (n = t->next) ) {
+	if ( !(n->flags & CHECKED) )
+	    checkline(n);
 
-    /* ANY leading `#`'s make this into an ETX header
-     */
-    if ( i ) {
-	*htyp = ETX;
-	return 1;
-    }
-
-    /* then check for setext-style HEADER
-     *                             ======
-     */
-
-    if ( t->next ) {
-	char *q = T(t->next->text);
-
-	if ( (*q == '=') || (*q == '-') ) {
-	    for (i=1; i < S(t->next->text); i++)
-		if ( q[0] != q[i] )
-		    return 0;
+	if ( n->kind == chk_dash || n->kind == chk_equal ) {
 	    *htyp = SETEXT;
 	    return 1;
 	}
@@ -332,55 +388,138 @@ ishdr(Line *t, int *htyp)
 
 
 static int
-isdefinition(Line *t)
+ishdr(Line *t, int *htyp)
 {
-#if DL_TAG_EXTENSION
-    return t && t->next
-	     && (S(t->text) > 2)
-	     && (t->dle == 0)
-	     && (T(t->text)[0] == '=')
-	     && (T(t->text)[S(t->text)-1] == '=')
-	     && ( (t->next->dle >= 4) || isdefinition(t->next) );
-#else
-    return 0;
+    /* ANY leading `#`'s make this into an ETX header
+     */
+    if ( (t->dle == 0) && (S(t->text) > 1) && (T(t->text)[0] == '#') ) {
+	*htyp = ETX;
+	return 1;
+    }
+
+    /* And if not, maybe it's a SETEXT header instead
+     */
+    return issetext(t, htyp);
+}
+
+
+static inline int
+end_of_block(Line *t)
+{
+    int dummy;
+    
+    if ( !t )
+	return 0;
+	
+    return ( (S(t->text) <= t->dle) || ishr(t) || ishdr(t, &dummy) );
+}
+
+
+static Line*
+is_discount_dt(Line *t, int *clip)
+{
+#if USE_DISCOUNT_DL
+    if ( t && t->next
+	   && (S(t->text) > 2)
+	   && (t->dle == 0)
+	   && (T(t->text)[0] == '=')
+	   && (T(t->text)[S(t->text)-1] == '=') ) {
+	if ( t->next->dle >= 4 ) {
+	    *clip = 4;
+	    return t;
+	}
+	else
+	    return is_discount_dt(t->next, clip);
+    }
 #endif
+    return 0;
 }
 
 
 static int
-islist(Line *t, int *trim)
+is_extra_dd(Line *t)
+{
+    return (t->dle < 4) && (T(t->text)[t->dle] == ':')
+			&& isspace(T(t->text)[t->dle+1]);
+}
+
+
+static Line*
+is_extra_dt(Line *t, int *clip)
+{
+#if USE_EXTRA_DL
+    
+    if ( t && t->next && T(t->text)[0] != '='
+		      && T(t->text)[S(t->text)-1] != '=') {
+	Line *x;
+    
+	if ( iscode(t) || end_of_block(t) )
+	    return 0;
+
+	if ( (x = skipempty(t->next)) && is_extra_dd(x) ) {
+	    *clip = x->dle+2;
+	    return t;
+	}
+	
+	if ( x=is_extra_dt(t->next, clip) )
+	    return x;
+    }
+#endif
+    return 0;
+}
+
+
+static Line*
+isdefinition(Line *t, int *clip, int *kind)
+{
+    Line *ret;
+
+    *kind = 1;
+    if ( ret = is_discount_dt(t,clip) )
+	return ret;
+
+    *kind=2;
+    return is_extra_dt(t,clip);
+}
+
+
+static int
+islist(Line *t, int *clip, DWORD flags, int *list_type)
 {
     int i, j;
     char *q;
     
-    if ( iscode(t) || blankline(t) || ishdr(t,&i) || ishr(t) )
+    if ( end_of_block(t) )
 	return 0;
 
-    if ( isdefinition(t) ) {
-	*trim = 4;
+    if ( !(flags & (MKD_NODLIST|MKD_STRICT)) && isdefinition(t,clip,list_type) )
 	return DL;
-    }
 
     if ( strchr("*-+", T(t->text)[t->dle]) && isspace(T(t->text)[t->dle+1]) ) {
 	i = nextnonblank(t, t->dle+1);
-	*trim = (i > 4) ? 4 : i;
-	return UL;
+	*clip = (i > 4) ? 4 : i;
+	*list_type = UL;
+	return AL;
     }
 
     if ( (j = nextblank(t,t->dle)) > t->dle ) {
 	if ( T(t->text)[j-1] == '.' ) {
-#if ALPHA_LIST
-	if ( (j == t->dle + 2) && isalpha(T(t->text)[t->dle]) ) {
-	    j = nextnonblank(t,j);
-	    *trim = j;
-	    return AL;
-	}
-#endif
+
+	    if ( !(flags & (MKD_NOALPHALIST|MKD_STRICT))
+				    && (j == t->dle + 2)
+			  && isalpha(T(t->text)[t->dle]) ) {
+		j = nextnonblank(t,j);
+		*clip = (j > 4) ? 4 : j;
+		*list_type = AL;
+		return AL;
+	    }
+
 	    strtoul(T(t->text)+t->dle, &q, 10);
 	    if ( (q > T(t->text)+t->dle) && (q == T(t->text) + (j-1)) ) {
 		j = nextnonblank(t,j);
-		*trim = j;
-		return OL;
+		*clip = (j > 4) ? 4 : j;
+		*list_type = OL;
+		return AL;
 	    }
 	}
     }
@@ -411,7 +550,8 @@ headerblock(Paragraph *pp, int htyp)
 	     * the leading and trailing `#`'s
 	     */
 
-	    for (i=0; (T(p->text)[i] == T(p->text)[0]) && (i < S(p->text)-1); i++)
+	    for (i=0; (T(p->text)[i] == T(p->text)[0]) && (i < S(p->text)-1)
+						       && (i < 6); i++)
 		;
 
 	    pp->hnumber = i;
@@ -420,6 +560,7 @@ headerblock(Paragraph *pp, int htyp)
 		++i;
 
 	    CLIP(p->text, 0, i);
+	    UNCHECK(p);
 
 	    for (j=S(p->text); (j > 1) && (T(p->text)[j-1] == '#'); --j)
 		;
@@ -456,6 +597,46 @@ codeblock(Paragraph *p)
 }
 
 
+#ifdef WITH_FENCED_CODE
+static int
+iscodefence(Line *r, int size)
+{
+    if ( !(r->flags & CHECKED) )
+	checkline(r);
+
+    return (r->kind == chk_tilde) && (r->count >= size);
+}
+
+static Paragraph *
+fencedcodeblock(ParagraphRoot *d, Line **ptr)
+{
+    Line *first, *r;
+    Paragraph *ret;
+
+    first = (*ptr);
+    
+    /* don't allow zero-length code fences
+     */
+    if ( (first->next == 0) || iscodefence(first->next, first->count) )
+	return 0;
+
+    /* find the closing fence, discard the fences,
+     * return a Paragraph with the contents
+     */
+    for ( r = first; r && r->next; r = r->next )
+	if ( iscodefence(r->next, first->count) ) {
+	    (*ptr) = r->next->next;
+	    ret = Pp(d, first->next, CODE);
+	    ___mkd_freeLine(first);
+	    ___mkd_freeLine(r->next);
+	    r->next = 0;
+	    return ret;
+	}
+    return 0;
+}
+#endif
+
+
 static int
 centered(Line *first, Line *last)
 {
@@ -475,27 +656,32 @@ centered(Line *first, Line *last)
 
 
 static int
-endoftextblock(Line *t, int toplevelblock)
+endoftextblock(Line *t, int toplevelblock, DWORD flags)
 {
     int z;
 
-    if ( blankline(t)||isquote(t)||iscode(t)||ishdr(t,&z)||ishr(t) )
+    if ( end_of_block(t) || isquote(t) )
 	return 1;
 
-    /* HORRIBLE STANDARDS KLUDGE: Toplevel paragraphs eat absorb adjacent
-     * list items, but sublevel blocks behave properly.
+    /* HORRIBLE STANDARDS KLUDGES:
+     * 1. non-toplevel paragraphs absorb adjacent code blocks
+     * 2. Toplevel paragraphs eat absorb adjacent list items,
+     *    but sublevel blocks behave properly.
+     * (What this means is that we only need to check for code
+     *  blocks at toplevel, and only check for list items at
+     *  nested levels.)
      */
-    return toplevelblock ? 0 : islist(t,&z);
+    return toplevelblock ? 0 : islist(t,&z,flags,&z);
 }
 
 
 static Line *
-textblock(Paragraph *p, int toplevel)
+textblock(Paragraph *p, int toplevel, DWORD flags)
 {
     Line *t, *next;
 
     for ( t = p->text; t ; t = next ) {
-	if ( ((next = t->next) == 0) || endoftextblock(next, toplevel) ) {
+	if ( ((next = t->next) == 0) || endoftextblock(next, toplevel, flags) ) {
 	    p->align = centered(p->text, t);
 	    t->next = 0;
 	    return next;
@@ -522,27 +708,32 @@ szmarkerclass(char *p)
  * check if the first line of a quoted block is the special div-not-quote
  * marker %[kind:]name%
  */
-static int
-isdivmarker(Line *p)
-{
-#if DIV_QUOTE
-    char *s = T(p->text);
-    int len = S(p->text);
-    int i;
+#define iscsschar(c) (isalpha(c) || (c == '-') || (c == '_') )
 
-    if ( !(len && s[0] == '%' && s[len-1] == '%') ) return 0;
+static int
+isdivmarker(Line *p, int start, DWORD flags)
+{
+    char *s;
+    int last, i;
+
+    if ( flags & (MKD_NODIVQUOTE|MKD_STRICT) )
+	return 0;
+
+    last= S(p->text) - (1 + start);
+    s   = T(p->text) + start;
+
+    if ( (last <= 0) || (*s != '%') || (s[last] != '%') )
+	return 0;
 
     i = szmarkerclass(s+1);
-    --len;
 
-    while ( ++i < len )
-	if ( !isalnum(s[i]) )
+    if ( !iscsschar(s[i+1]) )
+	return 0;
+    while ( ++i < last )
+	if ( !(isdigit(s[i]) || iscsschar(s[i])) )
 	    return 0;
 
     return 1;
-#else
-    return 0;
-#endif
 }
 
 
@@ -551,32 +742,44 @@ isdivmarker(Line *p)
  *
  * one sick horrible thing about blockquotes is that even though
  * it just takes ^> to start a quote, following lines, if quoted,
- * assume that the prefix is ``>''.   This means that code needs
+ * assume that the prefix is ``> ''.   This means that code needs
  * to be indented *5* spaces from the leading '>', but *4* spaces
  * from the start of the line.   This does not appear to be 
  * documented in the reference implementation, but it's the
  * way the markdown sample web form at Daring Fireball works.
  */
 static Line *
-quoteblock(Paragraph *p)
+quoteblock(Paragraph *p, DWORD flags)
 {
     Line *t, *q;
     int qp;
 
     for ( t = p->text; t ; t = q ) {
 	if ( isquote(t) ) {
-	    qp = (T(t->text)[1] == ' ') ? 2 : 1;
+	    /* clip leading spaces */
+	    for (qp = 0; T(t->text)[qp] != '>'; qp ++)
+		/* assert: the first nonblank character on this line
+		 * will be a >
+		 */;
+	    /* clip '>' */
+	    qp++;
+	    /* clip next space, if any */
+	    if ( T(t->text)[qp] == ' ' )
+		qp++;
 	    CLIP(t->text, 0, qp);
+	    UNCHECK(t);
 	    t->dle = mkd_firstnonblank(t);
 	}
 
-	if ( !(q = skipempty(t->next)) || ((q != t->next) && !isquote(q)) ) {
+	q = skipempty(t->next);
+
+	if ( (q == 0) || ((q != t->next) && (!isquote(q) || isdivmarker(q,1,flags))) ) {
 	    ___mkd_freeLineRange(t, q);
 	    t = q;
 	    break;
 	}
     }
-    if ( isdivmarker(p->text) ) {
+    if ( isdivmarker(p->text,0,flags) ) {
 	char *prefix = "class";
 	int i;
 	
@@ -587,7 +790,7 @@ quoteblock(Paragraph *p)
 	    /* and this would be an "%id:" prefix */
 	    prefix="id";
 	    
-	if ( p->ident = malloc(4+i+S(q->text)) )
+	if ( p->ident = malloc(4+strlen(prefix)+S(q->text)) )
 	    sprintf(p->ident, "%s=\"%.*s\"", prefix, S(q->text)-(i+2),
 						     T(q->text)+(i+1) );
 
@@ -597,8 +800,7 @@ quoteblock(Paragraph *p)
 }
 
 
-static Paragraph *Pp(ParagraphRoot *, Line *, int);
-static Paragraph *compile(Line *, int, MMIOT *);
+typedef int (*linefn)(Line *);
 
 
 /*
@@ -608,7 +810,7 @@ static Paragraph *compile(Line *, int, MMIOT *);
  * marker, but multiple paragraphs need to start with a 4-space indent.
  */
 static Line *
-listitem(Paragraph *p, int indent)
+listitem(Paragraph *p, int indent, DWORD flags, linefn check)
 {
     Line *t, *q;
     int clip = indent;
@@ -616,6 +818,7 @@ listitem(Paragraph *p, int indent)
 
     for ( t = p->text; t ; t = q) {
 	CLIP(t->text, 0, clip);
+	UNCHECK(t);
 	t->dle = mkd_firstnonblank(t);
 
 	if ( (q = skipempty(t->next)) == 0 ) {
@@ -624,19 +827,24 @@ listitem(Paragraph *p, int indent)
 	}
 
 	/* after a blank line, the next block needs to start with a line
-	 * that's indented 4 spaces, but after that the line doesn't
-	 * need any indentation
+	 * that's indented 4(? -- reference implementation allows a 1
+	 * character indent, but that has unfortunate side effects here)
+	 * spaces, but after that the line doesn't need any indentation
 	 */
 	if ( q != t->next ) {
-	    if (q->dle < 4) {
+	    if (q->dle < indent) {
 		q = t->next;
 		t->next = 0;
 		return q;
 	    }
-	    indent = 4;
+	    /* indent at least 2, and at most as
+	     * as far as the initial line was indented. */
+	    indent = clip ? clip : 2;
 	}
 
-	if ( (q->dle < indent) && (ishr(q) || islist(q,&z)) && !ishdr(q,&z) ) {
+	if ( (q->dle < indent) && (ishr(q) || islist(q,&z,flags,&z)
+					   || (check && (*check)(q)))
+			       && !issetext(q,&z) ) {
 	    q = t->next;
 	    t->next = 0;
 	    return q;
@@ -649,37 +857,82 @@ listitem(Paragraph *p, int indent)
 
 
 static Line *
-listblock(Paragraph *top, int trim, MMIOT *f)
+definition_block(Paragraph *top, int clip, MMIOT *f, int kind)
+{
+    ParagraphRoot d = { 0, 0 };
+    Paragraph *p;
+    Line *q = top->text, *text = 0, *labels; 
+    int z, para;
+
+    while (( labels = q )) {
+
+	if ( (q = isdefinition(labels, &z, &kind)) == 0 )
+	    break;
+
+	if ( (text = skipempty(q->next)) == 0 )
+	    break;
+
+	if (( para = (text != q->next) ))
+	    ___mkd_freeLineRange(q, text);
+	
+	q->next = 0; 
+	if ( kind == 1 /* discount dl */ )
+	    for ( q = labels; q; q = q->next ) {
+		CLIP(q->text, 0, 1);
+		UNCHECK(q);
+		S(q->text)--;
+	    }
+
+    dd_block:
+	p = Pp(&d, text, LISTITEM);
+
+	text = listitem(p, clip, f->flags, (kind==2) ? is_extra_dd : 0);
+	p->down = compile(p->text, 0, f);
+	p->text = labels; labels = 0;
+
+	if ( para && p->down ) p->down->align = PARA;
+
+	if ( (q = skipempty(text)) == 0 )
+	    break;
+
+	if (( para = (q != text) )) {
+	    Line anchor;
+
+	    anchor.next = text;
+	    ___mkd_freeLineRange(&anchor,q);
+	    text = q;
+	    
+	}
+
+	if ( kind == 2 && is_extra_dd(q) )
+	    goto dd_block;
+    }
+    top->text = 0;
+    top->down = T(d);
+    return text;
+}
+
+
+static Line *
+enumerated_block(Paragraph *top, int clip, MMIOT *f, int list_class)
 {
     ParagraphRoot d = { 0, 0 };
     Paragraph *p;
     Line *q = top->text, *text;
-    Line *label;
-    int para = 0;
+    int para = 0, z;
 
     while (( text = q )) {
-	if ( top->typ == DL ) {
-	    Line *lp;
-
-	    for ( lp = label = text; lp ; lp = lp->next ) {
-		text = lp->next;
-		CLIP(lp->text, 0, 1);
-		S(lp->text)--;
-		if ( !isdefinition(lp->next) )
-		    lp->next = 0;
-	    }
-	}
-	else label = 0;
-
+	
 	p = Pp(&d, text, LISTITEM);
-	text = listitem(p, trim);
+	text = listitem(p, clip, f->flags, 0);
 
 	p->down = compile(p->text, 0, f);
-	p->text = label;
+	p->text = 0;
 
-	if ( para && (top->typ != DL) && p->down ) p->down->align = PARA;
+	if ( para && p->down ) p->down->align = PARA;
 
-	if ( !(q = skipempty(text)) || (islist(q, &trim) == 0) )
+	if ( (q = skipempty(text)) == 0
+			     || islist(q, &clip, f->flags, &z) != list_class )
 	    break;
 
 	if ( para = (q != text) ) {
@@ -687,9 +940,9 @@ listblock(Paragraph *top, int trim, MMIOT *f)
 
 	    anchor.next = text;
 	    ___mkd_freeLineRange(&anchor, q);
-	}
 
-	if ( para && (top->typ != DL) && p->down ) p->down->align = PARA;
+	    if ( p->down ) p->down->align = PARA;
+	}
     }
     top->text = 0;
     top->down = T(d);
@@ -724,7 +977,7 @@ addfootnote(Line *p, MMIOT* f)
     CREATE(foot->tag);
     CREATE(foot->link);
     CREATE(foot->title);
-    foot->height = foot->width = 0;
+    foot->flags = foot->height = foot->width = 0;
 
     for (j=i=p->dle+1; T(p->text)[j] != ']'; j++)
 	EXPAND(foot->tag) = T(p->text)[j];
@@ -733,6 +986,12 @@ addfootnote(Line *p, MMIOT* f)
     S(foot->tag)--;
     j = nextnonblank(p, j+2);
 
+    if ( (f->flags & MKD_EXTRA_FOOTNOTE) && (T(foot->tag)[0] == '^') ) {
+	while ( j < S(p->text) )
+	    EXPAND(foot->title) = T(p->text)[j++];
+	goto skip_to_end;
+    }
+
     while ( (j < S(p->text)) && !isspace(T(p->text)[j]) )
 	EXPAND(foot->link) = T(p->text)[j++];
     EXPAND(foot->link) = 0;
@@ -772,6 +1031,7 @@ addfootnote(Line *p, MMIOT* f)
 	--S(foot->title);
     }
 
+skip_to_end:
     ___mkd_freeLine(p);
     return np;
 }
@@ -809,6 +1069,118 @@ consume(Line *ptr, int *eaten)
 }
 
 
+/*
+ * top-level compilation; break the document into
+ * style, html, and source blocks with footnote links
+ * weeded out.
+ */
+static Paragraph *
+compile_document(Line *ptr, MMIOT *f)
+{
+    ParagraphRoot d = { 0, 0 };
+    ANCHOR(Line) source = { 0, 0 };
+    Paragraph *p = 0;
+    struct kw *tag;
+    int eaten, unclosed;
+
+    while ( ptr ) {
+	if ( !(f->flags & MKD_NOHTML) && (tag = isopentag(ptr)) ) {
+	    /* If we encounter a html/style block, compile and save all
+	     * of the cached source BEFORE processing the html/style.
+	     */
+	    if ( T(source) ) {
+		E(source)->next = 0;
+		p = Pp(&d, 0, SOURCE);
+		p->down = compile(T(source), 1, f);
+		T(source) = E(source) = 0;
+	    }
+	    p = Pp(&d, ptr, strcmp(tag->id, "STYLE") == 0 ? STYLE : HTML);
+	    ptr = htmlblock(p, tag, &unclosed);
+	    if ( unclosed ) {
+		p->typ = SOURCE;
+		p->down = compile(p->text, 1, f);
+		p->text = 0;
+	    }
+	}
+	else if ( isfootnote(ptr) ) {
+	    /* footnotes, like cats, sleep anywhere; pull them
+	     * out of the input stream and file them away for
+	     * later processing
+	     */
+	    ptr = consume(addfootnote(ptr, f), &eaten);
+	}
+	else {
+	    /* source; cache it up to wait for eof or the
+	     * next html/style block
+	     */
+	    ATTACH(source,ptr);
+	    ptr = ptr->next;
+	}
+    }
+    if ( T(source) ) {
+	/* if there's any cached source at EOF, compile
+	 * it now.
+	 */
+	E(source)->next = 0;
+	p = Pp(&d, 0, SOURCE);
+	p->down = compile(T(source), 1, f);
+    }
+    return T(d);
+}
+
+
+static int
+first_nonblank_before(Line *j, int dle)
+{
+    return (j->dle < dle) ? j->dle : dle;
+}
+
+
+static int
+actually_a_table(MMIOT *f, Line *pp)
+{
+    Line *r;
+    int j;
+    int c;
+
+    /* tables need to be turned on */
+    if ( f->flags & (MKD_STRICT|MKD_NOTABLES) )
+	return 0;
+
+    /* tables need three lines */
+    if ( !(pp && pp->next && pp->next->next) ) {
+	return 0;
+    }
+
+    /* all lines must contain |'s */
+    for (r = pp; r; r = r->next )
+	if ( !(r->flags & PIPECHAR) ) {
+	    return 0;
+	}
+
+    /* if the header has a leading |, all lines must have leading |'s */
+    if ( T(pp->text)[pp->dle] == '|' ) {
+	for ( r = pp; r; r = r->next )
+	    if ( T(r->text)[first_nonblank_before(r,pp->dle)] != '|' ) {
+		return 0;
+	    }
+    }
+
+    /* second line must be only whitespace, -, |, or : */
+    r = pp->next;
+
+    for ( j=r->dle; j < S(r->text); ++j ) {
+	c = T(r->text)[j];
+
+	if ( !(isspace(c)||(c=='-')||(c==':')||(c=='|')) ) {
+	    return 0;
+	}
+    }
+
+    return 1;
+}
+
+
 /*
  * break a collection of markdown input into
  * blocks of lists, code, html, and text to
@@ -819,46 +1191,49 @@ compile(Line *ptr, int toplevel, MMIOT *f)
 {
     ParagraphRoot d = { 0, 0 };
     Paragraph *p = 0;
-    char *key;
     Line *r;
     int para = toplevel;
-    int hdr_type, list_type, indent;
+    int blocks = 0;
+    int hdr_type, list_type, list_class, indent;
 
     ptr = consume(ptr, &para);
 
     while ( ptr ) {
-	if ( toplevel && !(f->flags & DENY_HTML) && (key = isopentag(ptr)) ) {
-	    p = Pp(&d, ptr, strcmp(key, "STYLE") == 0 ? STYLE : HTML);
-	    if ( strcmp(key, "!--") == 0 )
-		ptr = comment(p, key);
-	    else
-		ptr = htmlblock(p, key);
-	}
-	else if ( iscode(ptr) ) {
+	if ( iscode(ptr) ) {
 	    p = Pp(&d, ptr, CODE);
 	    
 	    if ( f->flags & MKD_1_COMPAT) {
 		/* HORRIBLE STANDARDS KLUDGE: the first line of every block
 		 * has trailing whitespace trimmed off.
 		 */
-		___mkd_tidy(p->text);
+		___mkd_tidy(&p->text->text);
 	    }
 	    
 	    ptr = codeblock(p);
 	}
+#if WITH_FENCED_CODE
+	else if ( iscodefence(ptr,3) && (p=fencedcodeblock(&d, &ptr)) )
+	    /* yay, it's already done */ ;
+#endif
 	else if ( ishr(ptr) ) {
 	    p = Pp(&d, 0, HR);
 	    r = ptr;
 	    ptr = ptr->next;
 	    ___mkd_freeLine(r);
 	}
-	else if (( list_type = islist(ptr, &indent) )) {
-	    p = Pp(&d, ptr, list_type);
-	    ptr = listblock(p, indent, f);
+	else if (( list_class = islist(ptr, &indent, f->flags, &list_type) )) {
+	    if ( list_class == DL ) {
+		p = Pp(&d, ptr, DL);
+		ptr = definition_block(p, indent, f, list_type);
+	    }
+	    else {
+		p = Pp(&d, ptr, list_type);
+		ptr = enumerated_block(p, indent, f, list_class);
+	    }
 	}
 	else if ( isquote(ptr) ) {
 	    p = Pp(&d, ptr, QUOTE);
-	    ptr = quoteblock(p);
+	    ptr = quoteblock(p, f->flags);
 	    p->down = compile(p->text, 1, f);
 	    p->text = 0;
 	}
@@ -866,19 +1241,19 @@ compile(Line *ptr, int toplevel, MMIOT *f)
 	    p = Pp(&d, ptr, HDR);
 	    ptr = headerblock(p, hdr_type);
 	}
-	else if ( toplevel && (isfootnote(ptr)) ) {
-	    ptr = consume(addfootnote(ptr, f), &para);
-	    continue;
-	}
 	else {
 	    p = Pp(&d, ptr, MARKUP);
-	    ptr = textblock(p, toplevel);
+	    ptr = textblock(p, toplevel, f->flags);
+	    /* tables are a special kind of paragraph */
+	    if ( actually_a_table(f, p->text) )
+		p->typ = TABLE;
 	}
 
 	if ( (para||toplevel) && !p->align )
 	    p->align = PARA;
 
-	para = toplevel;
+	blocks++;
+	para = toplevel || (blocks > 1);
 	ptr = consume(ptr, &para);
 
 	if ( para && !p->align )
@@ -889,19 +1264,6 @@ compile(Line *ptr, int toplevel, MMIOT *f)
 }
 
 
-static void
-initialize()
-{
-    static int first = 1;
-
-    if ( first-- > 0 ) {
-	first = 0;
-	INITRNG(time(0));
-	qsort(blocktags, SZTAGS, sizeof blocktags[0], (stfu)casort);
-    }
-}
-
-
 /*
  * the guts of the markdown() function, ripped out so I can do
  * debugging.
@@ -911,7 +1273,7 @@ initialize()
  * prepare and compile `text`, returning a Paragraph tree.
  */
 int
-mkd_compile(Document *doc, int flags)
+mkd_compile(Document *doc, DWORD flags)
 {
     if ( !doc )
 	return 0;
@@ -921,15 +1283,16 @@ mkd_compile(Document *doc, int flags)
 
     doc->compiled = 1;
     memset(doc->ctx, 0, sizeof(MMIOT) );
-    doc->ctx->flags = flags & USER_FLAGS;
-    doc->ctx->base = doc->base;
+    doc->ctx->ref_prefix= doc->ref_prefix;
+    doc->ctx->cb        = &(doc->cb);
+    doc->ctx->flags     = flags & USER_FLAGS;
     CREATE(doc->ctx->in);
     doc->ctx->footnotes = malloc(sizeof doc->ctx->footnotes[0]);
     CREATE(*doc->ctx->footnotes);
 
-    initialize();
+    mkd_initialize();
 
-    doc->code = compile(T(doc->content), 1, doc->ctx);
+    doc->code = compile_document(T(doc->content), doc->ctx);
     qsort(T(*doc->ctx->footnotes), S(*doc->ctx->footnotes),
 		        sizeof T(*doc->ctx->footnotes)[0],
 			           (stfu)__mkd_footsort);
